import re
from functools import wraps
from inspect import isawaitable
from typing import Any, Callable, List

from graphql import parse


def gql(value: str) -> str:
    parse(value)
    return value


# Adapted from this response in Stackoverflow
# http://stackoverflow.com/a/19053800/1072990
def to_camel_case(snake_str: str) -> str:
    components = snake_str.split("_")
    # We capitalize the first letter of each component except the first one
    # with the 'capitalize' method and join them together.
    return components[0] + "".join(x.capitalize() if x else "_" for x in components[1:])


def snake_argument(func: Callable) -> Callable:
    @wraps(func)
    async def wrap(*args, **kwargs):
        kwargs = {to_snake_case(k): v for k, v in kwargs.items()}
        return await func(*args, **kwargs)

    return wrap


def recursive_to_camel_case(d: Any) -> Any:
    if isinstance(d, list):
        return [recursive_to_camel_case(v) for v in d]
    if not isinstance(d, dict):
        return d

    _d = {}
    for k, v in d.items():
        _d[to_camel_case(k)] = recursive_to_camel_case(v)
    return _d


# From this response in Stackoverflow
# http://stackoverflow.com/a/1176023/1072990
def to_snake_case(name: str) -> str:
    s1 = re.sub(r'(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub(r'([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def recursive_to_snake_case(d: Any) -> Any:
    if isinstance(d, list):
        return [recursive_to_snake_case(v) for v in d]
    if not isinstance(d, dict):
        return d

    _d = {}
    for k, v in d.items():
        _d[to_snake_case(k)] = recursive_to_snake_case(v)
    return _d


def to_const(string: str) -> str:
    return re.sub(r'[\W|^]+', '_', string).upper()


def add_file_to_operations(operations, file_obj, path):
    """Handles the recursive algorithm for adding a file to the operations
    object"""
    if not path:
        if operations is not None:
            raise ValueError('Path in map does not lead to a null value')
        return file_obj
    if isinstance(operations, dict):
        key = path[0]
        sub_dict = add_file_to_operations(operations[key], file_obj, path[1:])
        return new_merged_dict(operations, {key: sub_dict})
    if isinstance(operations, list):
        index = int(path[0])
        sub_item = add_file_to_operations(operations[index], file_obj, path[1:])
        return new_list_with_replaced_item(operations, index, sub_item)
    raise TypeError('Operations must be a dict or a list of dicts')


def new_merged_dict(*dicts):
    """Merges dictionaries into a new dictionary. Necessary for python2 and
    python34 since neither have PEP448 implemented."""
    # Necessary for python2 support
    output = {}
    for d in dicts:
        output.update(d)
    return output


def new_list_with_replaced_item(input_list, index, new_value):
    """Creates new list with replaced item at specified index"""
    output = [i for i in input_list]
    output[index] = new_value
    return output


def place_files_in_operations(operations, files_map, form):
    """Replaces None placeholders in operations with file objects in the files
    dictionary, by following the files_map logic as specified within the 'map'
    request parameter in the multipart request spec"""
    path_to_key_iter = (
        (value.split('.'), key) for (key, values) in files_map.items() for value in values
    )
    # Since add_files_to_operations returns a new dict/list, first define
    # output to be operations itself
    output = operations
    for path, key in path_to_key_iter:
        file_obj = form.get(key)
        output = add_file_to_operations(output, file_obj, path)
    return output


async def execute_async_function(func, *args, **kwargs):
    result = func(*args, **kwargs)
    if isawaitable(result):
        result = await result
    return result


def join_type_defs(type_defs: List[str]) -> str:
    return "\n\n".join(t.strip() for t in type_defs)



import traceback
from collections import defaultdict
from enum import Enum
from functools import partial, wraps
from inspect import iscoroutinefunction, isfunction
from typing import Any, Callable, Dict, Mapping, Union

from graphql import (
    GraphQLFieldResolver,
    GraphQLResolveInfo,
    GraphQLSchema,
    GraphQLTypeResolver,
    assert_interface_type,
    assert_object_type,
    assert_union_type,
    is_interface_type,
    is_object_type,
    is_union_type,
)

from .utils import execute_async_function, recursive_to_snake_case, to_camel_case, to_snake_case

ReferenceResolver = Callable[[Any, GraphQLResolveInfo, dict], Any]
FieldResolverMap = Dict[str, Dict[str, GraphQLFieldResolver]]
TypeResolverMap = Dict[str, GraphQLTypeResolver]
ReferenceResolverMap = Dict[str, ReferenceResolver]

field_resolver_map: FieldResolverMap = defaultdict(dict)
type_resolver_map: TypeResolverMap = {}
reference_resolver_map: ReferenceResolverMap = {}


def reference_resolver(type_name: str):
    def wrap(func: ReferenceResolver):
        @wraps(func)
        def sync_resolver(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as exc:
                traceback.print_exc()
                raise exc

        @wraps(func)
        async def async_resolver(*args, **kwargs):
            try:
                return await execute_async_function(func, *args, **kwargs)
            except Exception as exc:
                traceback.print_exc()
                raise exc

        if iscoroutinefunction(func):
            reference_resolver_map[type_name] = async_resolver
            return async_resolver
        reference_resolver_map[type_name] = sync_resolver
        return sync_resolver

    return wrap


def type_resolver(type_name: str):
    def wrap(func):
        type_resolver_map[type_name] = func
        return func

    return wrap


def field_resolver(
    type_name: str,
    func_or_field: Union[GraphQLFieldResolver, str] = None,
    print_exc: bool = True,
    snake_argument: bool = True,
):
    def wrap(func: GraphQLFieldResolver):
        @wraps(func)
        def sync_resolver(*args, **kwargs):
            if snake_argument:
                kwargs = recursive_to_snake_case(kwargs)
            if not print_exc:
                return func(*args, **kwargs)

            try:
                return func(*args, **kwargs)
            except Exception as exc:
                traceback.print_exc()
                raise exc

        @wraps(func)
        async def async_resolver(*args, **kwargs):
            if snake_argument:
                kwargs = recursive_to_snake_case(kwargs)
            if not print_exc:
                return await execute_async_function(func, *args, **kwargs)

            try:
                return await execute_async_function(func, *args, **kwargs)
            except Exception as exc:
                traceback.print_exc()
                raise exc

        if isinstance(func_or_field, str):
            name = to_camel_case(func_or_field or func.__name__)
        else:
            name = to_camel_case(func.__name__)

        if iscoroutinefunction(func):
            field_resolver_map[type_name][name] = async_resolver
            return async_resolver

        field_resolver_map[type_name][name] = sync_resolver
        return sync_resolver

    if isfunction(func_or_field):
        return wrap(func_or_field)

    return wrap


mutate = partial(field_resolver, 'Mutation')
query = partial(field_resolver, 'Query')
subscribe = partial(field_resolver, 'Subscription')


def register_reference_resolvers(schema: GraphQLSchema):
    for type_name, resolver in reference_resolver_map.items():
        type_ = schema.get_type(type_name)
        type_.__resolve_reference__ = resolver


def register_type_resolvers(schema: GraphQLSchema):
    for type_name, type_resolver in type_resolver_map.items():
        type_ = schema.get_type(type_name)
        if is_interface_type(type_):
            type_ = assert_interface_type(type_)
        elif is_union_type(type_):
            type_ = assert_union_type(type_)
        else:
            continue
        type_.resolve_type = type_resolver


def register_field_resolvers(schema: GraphQLSchema):
    for type_name, field_resolvers in field_resolver_map.items():
        type_ = schema.get_type(type_name)
        if is_object_type(type_):
            type_ = assert_object_type(type_)
        elif is_interface_type(type_):
            type_ = assert_interface_type(type_)
        else:
            continue

        for name, field_resolver in field_resolvers.items():
            field = type_.fields.get(name)
            if not field:
                continue
            if type_name == 'Subscription':
                field.subscribe = field_resolver
            else:
                field.resolve = field_resolver


def register_resolvers(schema: GraphQLSchema):
    register_field_resolvers(schema)
    register_type_resolvers(schema)
    register_reference_resolvers(schema)


def get_field_value(source, field_name):
    return (
        source.get(field_name) if isinstance(source, Mapping) else getattr(source, field_name, None)
    )


def default_field_resolver(source, info, **args):
    """Default field resolver.

    If a resolve function is not given, then a default resolve behavior is used which
    takes the property of the source object of the same name as the field and returns
    it as the result, or if it's a function, returns the result of calling that function
    while passing along args and context.

    For dictionaries, the field names are used as keys, for all other objects they are
    used as attribute names.
    """
    # Ensure source is a value for which property access is acceptable.
    value = get_field_value(source, to_snake_case(info.field_name))
    if value is None:
        value = get_field_value(source, info.field_name)

    if callable(value):
        return value(info, **args)
    if isinstance(value, Enum):
        return value.value
    return value



import json
from datetime import datetime, timezone
from inspect import isclass
from numbers import Number
from typing import Any, Dict

from graphql import (
    INVALID,
    GraphQLSchema,
    IntValueNode,
    StringValueNode,
    assert_scalar_type,
    is_scalar_type,
)
from graphql.pyutils import inspect, is_integer

ScalarTypeMap = Dict[str, Any]

scalar_type_map: ScalarTypeMap = {}


def scalar_type(_cls):
    # if _cls is a class, use _cls name.
    if isclass(_cls):
        scalar_type_map[_cls.__name__] = _cls
        return _cls

    # if not, _cls is type name.
    def wrap(cls):
        scalar_type_map[_cls] = cls
        return cls

    return wrap


def register_scalars(schema: GraphQLSchema):
    for type_name, _scalar_type in scalar_type_map.items():
        type_ = schema.get_type(type_name)
        if not type_:
            continue
        if not is_scalar_type(type_):
            raise Exception(f'{type_name} is not a scalar type.')
        type_ = assert_scalar_type(type_)
        serialize = getattr(_scalar_type, 'serialize', None)
        if serialize:
            type_.serialize = serialize
        parse_value = getattr(_scalar_type, 'parse_value', None)
        if parse_value:
            type_.parse_value = parse_value
        parse_literal = getattr(_scalar_type, 'parse_literal', None)
        if parse_literal:
            type_.parse_literal = parse_literal


@scalar_type
class Upload:
    description = """The `Upload` type represents a type implement graphql multipart request spec
    (https://github.com/jaydenseric/graphql-multipart-request-spec)."""

    @staticmethod
    def serialize(value: Any) -> Any:
        return value

    @staticmethod
    def parse_value(value: Any) -> Any:
        return value

    @staticmethod
    def parse_literal(ast, _variables=None):
        return ast


@scalar_type
class Timestamp:
    description = "The `Timestamp` represents a millisecond unix timestamp with time zone."

    @staticmethod
    def serialize(value: Any) -> int:
        if isinstance(value, Number):
            return int(value)
        if isinstance(value, str) and value.isdigit():
            return int(value)

        if not isinstance(value, datetime):
            raise TypeError(f'Timestamp cannot represent non datetime value: {inspect(value)}')
        return int(value.timestamp() * 1000)

    @staticmethod
    def parse_value(value: Any) -> datetime:
        if not is_integer(value):
            raise TypeError(f'Timestamp cannot represent non datetime value: {inspect(value)}')
        return datetime.utcfromtimestamp(int(value) / 1000).replace(tzinfo=timezone.utc)

    @staticmethod
    def parse_literal(ast, _variables=None):
        if isinstance(ast, IntValueNode):
            return datetime.utcfromtimestamp(int(ast.value) / 1000).replace(tzinfo=timezone.utc)

        return INVALID


@scalar_type
class JSONString:
    description = "The `JSONString` represents a json string."

    @staticmethod
    def serialize(value: Any) -> str:
        if isinstance(value, (dict, list, tuple)):
            return json.dumps(value)

        if isinstance(value, str):
            return value
        if isinstance(value, bool):
            return "true" if value else "false"
        return str(value)

    @staticmethod
    def parse_value(value: Any) -> dict:
        if not isinstance(value, str):
            raise TypeError(f'JSONString cannot represent non string value: {inspect(value)}')
        return json.loads(value)

    @staticmethod
    def parse_literal(ast, _variables=None):
        if isinstance(ast, StringValueNode):
            return json.loads(ast.value)

        return INVALID



from typing import List, Union, cast

from graphql import (
    GraphQLObjectType,
    GraphQLSchema,
    GraphQLUnionType,
    build_schema,
    extend_schema,
    parse,
)

from .enum import register_enums
from .federation import (
    federation_entity_type_defs,
    federation_service_type_defs,
    get_entity_types,
    purge_schema_directives,
    remove_subscription,
    resolve_entities,
)
from .resolver import register_resolvers
from .scalar import register_scalars
from .utils import join_type_defs


def make_schema(
    type_defs: Union[str, List[str]],
    assume_valid: bool = False,
    assume_valid_sdl: bool = False,
    no_location: bool = False,
    experimental_fragment_variables: bool = False,
    federation: bool = False,
) -> GraphQLSchema:
    if isinstance(type_defs, list):
        type_defs = join_type_defs(type_defs)

    if not federation:
        schema = build_schema(
            type_defs, assume_valid, assume_valid_sdl, no_location, experimental_fragment_variables
        )
    else:
        schema = make_federation_schema(type_defs)

    register_resolvers(schema)
    register_enums(schema)
    register_scalars(schema)
    return schema


def make_schema_from_file(
    file: str,
    assume_valid: bool = False,
    assume_valid_sdl: bool = False,
    no_location: bool = False,
    experimental_fragment_variables: bool = False,
    federation: bool = False,
) -> GraphQLSchema:
    with open(file, 'r') as f:
        schema = make_schema(
            f.read(),
            assume_valid,
            assume_valid_sdl,
            no_location,
            experimental_fragment_variables,
            federation,
        )
        return schema


def make_federation_schema(
    type_defs: str,
    assume_valid: bool = False,
    assume_valid_sdl: bool = False,
    no_location: bool = False,
    experimental_fragment_variables: bool = False,
):
    # Remove custom schema directives (to avoid apollo-gateway crashes).
    sdl = purge_schema_directives(type_defs)

    # remove subscription because Apollo Federation not support subscription yet.
    sdl = remove_subscription(type_defs)

    type_defs = join_type_defs([type_defs, federation_service_type_defs])
    schema = build_schema(
        type_defs, assume_valid, assume_valid_sdl, no_location, experimental_fragment_variables
    )
    entity_types = get_entity_types(schema)
    if entity_types:
        schema = extend_schema(schema, parse(federation_entity_type_defs))

        # Add _entities query.
        entity_type = schema.get_type("_Entity")
        if entity_type:
            entity_type = cast(GraphQLUnionType, entity_type)
            entity_type.types = entity_types

        query_type = schema.get_type("Query")
        if query_type:
            query_type = cast(GraphQLObjectType, query_type)
            query_type.fields["_entities"].resolve = resolve_entities

    # Add _service query.
    query_type = schema.get_type("Query")
    if query_type:
        query_type = cast(GraphQLObjectType, query_type)
        query_type.fields["_service"].resolve = lambda _service, info: {"sdl": sdl}

    return schema



from dataclasses import dataclass, field
from typing import Dict, List, Optional, Set, Union

import graphql

from .utils import to_snake_case

GraphQLNode = Union[graphql.FieldNode, graphql.FragmentDefinitionNode]


@dataclass
class FieldMeta:
    name: str
    depth: int
    sections: List[str] = field(default_factory=list)
    sub_fields: Dict[str, 'FieldMeta'] = field(default_factory=dict)
    fragments: Set[str] = field(default_factory=set)
    inline_fragments: Dict[str, 'FieldMeta'] = field(default_factory=dict)

    def get_sub_field(self, name) -> Optional['FieldMeta']:
        return self.sub_fields.get(name)

    def add_section_or_sub_field(self, field_: 'FieldMeta') -> None:
        if not field_:
            return None

        if field_.sections or field_.sub_fields or field_.fragments:
            self.sub_fields[field_.name] = field_
        else:
            self.sections.append(field_.name)

    def replace_fragments(self, fragments: Dict[str, 'FieldMeta']) -> None:
        if not self.fragments and not self.sub_fields:
            return None

        for name in self.fragments:
            # In the real case, name in fragments surely.
            fragment = fragments[name]
            if fragment.fragments:
                fragment.replace_fragments(fragments)
            self.sections.extend(fragment.sections)

            for field_name, field_ in fragment.sub_fields.items():
                if field_.fragments:
                    field_.replace_fragments(fragments)
                self.sub_fields[field_name] = field_
        if self.depth > 1:
            for sub_field in self.sub_fields.values():
                if sub_field.fragments:
                    sub_field.replace_fragments(fragments)
        else:
            # clear sub fields if depth less than 1
            self.sub_fields = {}

        self.fragments = set()


def parse_fragment_nodes(nodes: Dict[str, graphql.FragmentDefinitionNode]) -> Dict[str, FieldMeta]:
    fragments = {name: parse_node(node, depth=2) for name, node in nodes.items()}
    for fragment in fragments.values():
        if fragment.fragments:
            fragment.replace_fragments(fragments)
    return fragments


def parse_node(node: GraphQLNode, depth: int = 2) -> Optional[FieldMeta]:
    if depth == 0:
        return None

    if isinstance(node, graphql.InlineFragmentNode):
        name = node.type_condition.name.value
    else:
        name = to_snake_case(node.name.value)
    field_meta = FieldMeta(name=name, depth=depth)
    if not node.selection_set:
        return field_meta

    depth -= 1
    for field_node in node.selection_set.selections:
        if isinstance(field_node, graphql.FragmentSpreadNode):
            field_meta.fragments.add(to_snake_case(field_node.name.value))
            continue

        if isinstance(field_node, graphql.InlineFragmentNode):
            meta = parse_node(field_node, depth=depth + 1)
            field_meta.inline_fragments[meta.name] = meta
            continue

        if not field_node.selection_set:
            name = field_node.name.value
            if not name.startswith('__'):
                field_meta.sections.append(to_snake_case(name))
            continue

        if depth == 0:
            continue

        field_meta.add_section_or_sub_field(parse_node(field_node, depth=depth))

    return field_meta


def parse_info(info: graphql.GraphQLResolveInfo, depth: int = 2, index: int = 0) -> 'FieldMeta':
    if 'fragments' not in info.context:
        fragments = parse_fragment_nodes(info.fragments)
        info.context['fragments'] = fragments
    else:
        fragments = info.context['fragments']

    field_metas = info.context.get('field_metas') or {}
    if index not in field_metas:
        field_meta = parse_node(info.field_nodes[index], depth=depth)
        field_meta.replace_fragments(fragments)
        field_metas[index] = field_meta

    return field_metas[index]


def fake_info(query) -> graphql.GraphQLResolveInfo:
    from unittest.mock import MagicMock

    doc = graphql.parse(query)
    return MagicMock(
        context={},
        field_nodes=doc.definitions[0].selection_set.selections,
        fragments={node.name.value: node for node in doc.definitions[1:]},
    )



import re
from inspect import isawaitable
from typing import Any, List

from graphql import (
    DirectiveNode,
    GraphQLInputObjectType,
    GraphQLNamedType,
    GraphQLObjectType,
    GraphQLResolveInfo,
    GraphQLSchema,
)

federation_service_type_defs = """
    scalar _Any

    type _Service {
      sdl: String
    }

    extend type Query {
      _service: _Service!
    }

    directive @external on FIELD_DEFINITION
    directive @requires(fields: String) on FIELD_DEFINITION
    directive @provides(fields: String) on FIELD_DEFINITION
    directive @key(fields: String) on OBJECT | INTERFACE

    # this is an optional directive discussed below
    directive @extends on OBJECT | INTERFACE
"""

federation_entity_type_defs = """
    # a union of all types that use the @key directive
    union _Entity

    extend type Query {
        _entities(representations: [_Any!]!): [_Entity]!
    }
"""

_i_token_delimiter = r"(?:^|[\s\r\n]+|$)"
_i_token_name = "[_A-Za-z][_0-9A-Za-z]*"
_i_token_arguments = r"\([^)]*\)"
_i_token_location = "[_A-Za-z][_0-9A-Za-z]*"

_r_directive_definition = re.compile(
    "("
    f"{_i_token_delimiter}directive"
    f"(?:{_i_token_delimiter})?@({_i_token_name})"
    f"(?:(?:{_i_token_delimiter})?{_i_token_arguments})?"
    f"{_i_token_delimiter}on"
    f"{_i_token_delimiter}(?:[|]{_i_token_delimiter})?{_i_token_location}"
    f"(?:{_i_token_delimiter}[|]{_i_token_delimiter}{_i_token_location})*"
    ")"
    f"(?={_i_token_delimiter})",
)

_r_directive = re.compile(
    "("
    f"(?:{_i_token_delimiter})?@({_i_token_name})"
    f"(?:(?:{_i_token_delimiter})?{_i_token_arguments})?"
    ")"
    f"(?={_i_token_delimiter})",
)

_allowed_directives = [
    "skip",  # Default directive as per specs.
    "include",  # Default directive as per specs.
    "deprecated",  # Default directive as per specs.
    "external",  # Federation directive.
    "requires",  # Federation directive.
    "provides",  # Federation directive.
    "key",  # Federation directive.
    "extends",  # Federation directive.
]

_r_subscription = re.compile("type\s+Subscription\s*{(.|\n)*}")  # noqa: W605


def purge_schema_directives(joined_type_defs: str) -> str:
    """Remove custom schema directives from federation."""
    joined_type_defs = _r_directive_definition.sub("", joined_type_defs)
    joined_type_defs = _r_directive.sub(
        lambda m: m.group(1) if m.group(2) in _allowed_directives else "", joined_type_defs,
    )
    return joined_type_defs


def remove_subscription(joined_type_defs: str) -> str:
    return _r_subscription.sub("", joined_type_defs)


def gather_directives(type_object: GraphQLNamedType,) -> List[DirectiveNode]:
    """Get all directive attached to a type."""
    directives: List[DirectiveNode] = []

    if hasattr(type_object, "extension_ast_nodes"):
        if type_object.extension_ast_nodes:
            for ast_node in type_object.extension_ast_nodes:
                if ast_node.directives:
                    directives.extend(ast_node.directives)

    if hasattr(type_object, "ast_node"):
        if type_object.ast_node and type_object.ast_node.directives:
            directives.extend(type_object.ast_node.directives)

    return directives


def includes_directive(type_object: GraphQLNamedType, directive_name: str,) -> bool:
    """Check if specified type includes a directive."""
    if isinstance(type_object, GraphQLInputObjectType):
        return False

    directives = gather_directives(type_object)
    return any([d.name.value == directive_name for d in directives])


def get_entity_types(schema: GraphQLSchema) -> List[GraphQLNamedType]:
    """Get all types that include the @key directive."""
    schema_types = schema.type_map.values()

    def check_type(t):
        return isinstance(t, GraphQLObjectType) and includes_directive(t, "key")

    return [t for t in schema_types if check_type(t)]


def resolve_entities(_: Any, info: GraphQLResolveInfo, **kwargs) -> Any:
    representations = list(kwargs.get("representations", list()))

    result = []
    for reference in representations:
        __typename = reference["__typename"]
        type_object = info.schema.get_type(__typename)

        if not type_object or not isinstance(type_object, GraphQLObjectType):
            raise Exception(
                f"The `_entities` resolver tried to load an entity for"
                f' type "{__typename}", but no object type of that name'
                f" was found in the schema",
            )

        resolve_reference = getattr(
            type_object, "__resolve_reference__", lambda o, i, r: reference,
        )

        representation = resolve_reference(type_object, info, reference)

        if isawaitable(representation):
            result.append(add_typename_to_async_return(representation, __typename))
        else:
            result.append(add_typename_to_possible_return(representation, __typename))

    return result


def add_typename_to_possible_return(obj: Any, typename: str) -> Any:
    if obj is not None:
        if isinstance(obj, dict):
            obj["__typename"] = typename
        else:
            setattr(obj, f"_{obj.__class__.__name__}__typename", typename)
        return obj
    return {"__typename": typename}


async def add_typename_to_async_return(obj: Any, typename: str) -> Any:
    return add_typename_to_possible_return(await obj, typename)



from enum import Enum
from inspect import isclass
from typing import Dict, Type

from graphql import GraphQLEnumValue, GraphQLSchema, assert_enum_type, is_enum_type

EnumTypeMap = Dict[str, Type[Enum]]

enum_type_map: EnumTypeMap = {}


def enum_type(_cls):
    # if _cls is a class, use _cls name.
    if isclass(_cls):
        enum_type_map[_cls.__name__] = _cls
        return _cls

    # if not, _cls is type name.
    def wrap(cls):
        if isinstance(cls, Enum):
            raise Exception('enum_resolver must resolve a Enum class.')
        enum_type_map[_cls] = cls
        return cls

    return wrap


def register_enums(schema: GraphQLSchema):
    for type_name, type_ in schema.type_map.items():
        if not is_enum_type(type_) or type_name.startswith('__'):
            continue

        type_ = assert_enum_type(type_)
        _enum_type = enum_type_map.get(type_name)
        if not _enum_type:
            type_.values = {value: GraphQLEnumValue(value) for value in type_.values}
        else:
            type_.values = {value.name: GraphQLEnumValue(value.value) for value in _enum_type}



from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, Union

# https://github.com/apollographql/subscriptions-transport-ws/blob/master/PROTOCOL.md
PROTOCOL = 'graphql-ws'


class MessageType(Enum):
    GQL_CONNECTION_INIT = 'connection_init'  # Client -> Server
    GQL_CONNECTION_ACK = 'connection_ack'  # Server -> Client
    GQL_CONNECTION_ERROR = 'connection_error'  # Server -> Client

    # NOTE: The keep alive message type does not follow the standard due to connection optimizations
    GQL_CONNECTION_KEEP_ALIVE = 'ka'  # Server -> Client
    GQL_CONNECTION_TERMINATE = 'connection_terminate'  # Client -> Server
    GQL_START = 'start'  # Client -> Server
    GQL_DATA = 'data'  # Server -> Client
    GQL_ERROR = 'error'  # Server -> Client
    GQL_COMPLETE = 'complete'  # Server -> Client
    GQL_STOP = 'stop'  # Client -> Server


@dataclass
class OperationMessagePayload:
    query: str = None
    variables: Dict[str, Any] = None
    operation_name: str = None

    @classmethod
    def build(cls, value: dict = None) -> 'OperationMessagePayload':
        if not value:
            return cls()
        return cls(
            query=value.get('query'),
            variables=value.get('variables'),
            operation_name=value.get('operationName'),
        )


@dataclass
class OperationMessage:
    type: MessageType

    id: str = None
    payload: Union[OperationMessagePayload, dict] = None

    @classmethod
    def build(cls, value: dict) -> 'OperationMessage':
        assert value is not None
        message = cls(type=MessageType(value.get('type')), id=value.get('id'),)
        payload = value.get('payload')
        if message.type == MessageType.GQL_START:
            payload = OperationMessagePayload.build(payload)
        message.payload = payload
        return message



from .enum import enum_type  # noqa
from .resolver import (  # noqa
    field_resolver,
    mutate,
    query,
    reference_resolver,
    subscribe,
    type_resolver,
)
from .scalar import scalar_type  # noqa
from .schema import make_schema, make_schema_from_file  # noqa
from .utils import gql  # noqa



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gql']

package_data = \
{'': ['*']}

install_requires = \
['graphql-core>=3,<4']

setup_kwargs = {
    'name': 'python-gql',
    'version': '0.1.2',
    'description': 'Python schema-first GraphQL library based on GraphQL-core',
    'long_description': '# python-gql\n\nPython schema-first GraphQL library based on GraphQL-core.\n\n\n## Requirements\n\nPython 3.7+\n\n## Installation\n\n`pip install python-gql`\n\n## Getting start\n\n```python\nimport graphql\nfrom gql import gql, make_schema, query, mutate\n\ntype_defs = gql("""\ntype Query {\n    hello(name: String!): String!\n}\n\ntype Post {\n    author: String!\n    comment: String!\n}\ntype Mutation {\n    addPost(author: String, comment: String): Post!\n}\n""")\n\n\n@query\ndef hello(parent, info, name: str) -> str:\n    return name\n\n\n@mutate\ndef add_post(parent, info, author: str = None, comment: str = None) -> dict:\n    return {\'author\': author, \'comment\': comment}\n\n\nschema = make_schema(type_defs)\n\nq = """\nquery {\n    hello(name: "graphql")\n}\n"""\nresult = graphql.graphql_sync(schema, q)\nprint(result.data)\n\n# result: {\'hello\': \'graphql\'}\n\nq = """\nmutation {\n    addPost(author: "syfun", comment: "This is a good library.") {\n        author\n        comment\n    }\n}\n"""\nresult = graphql.graphql_sync(schema, q)\nprint(result.data)\n\n# result: {\'addPost\': {\'author\': \'syfun\', \'comment\': \'This is a good library.\'}}\n```\n\n## Build schema\n\nThis library is `schema-first`, so you must build a schema explicitly.\n\nHere, we have two methods to build a schema, by `a type definitions` or `a schema file`.\n\n```python\nfrom gql import gql, make_schema\n\ntype_defs = gql("""\ntype Query {\n    hello(name: String!): String!\n}\n""")\n\nschema = make_schema(type_defs)\n```\n\n> `gql` function will check your type definitions syntax.\n\n```python\nfrom gql import make_schema_from_file\n\nschema = make_schema_from_file(\'./schema.graphql\')\n```\n\n## Resolver decorators\n\n> In Python, `decorator` is my favorite function, it save my life!\n\nWe can use `query`, `mutation`, `subscribe` to bind functions to GraphQL resolvers.\n\n```python\n@query\ndef hello(parent, info, name: str) -> str:\n    return name\n```\n\nThese decorators will auto convert the snake function to camel one.\n\n```python\n# add_port => addPost\n@mutate\ndef add_post(parent, info, author: str = None, comment: str = None) -> dict:\n    return {\'author\': author, \'comment\': comment}\n```\n\nWhen the funcation name different from the resolver name, you can give a name argument to these decorators.\n\n```python\n@query(\'hello\')\ndef hello_function(parent, info, name: str) -> str:\n    return name\n```\n\nAbout `subscribe`, please see [gql-subscriptions](gql-subscriptions).\n\n## Enum type decorator\n\nUse `enum_type` decorator with a python Enum class.\n\n```python\nfrom enum import Enum\n\nfrom gql import enum_type\n\n\n@enum_type\nclass Gender(Enum):\n    MALE = 1\n    FEMALE = 2\n```\n\n## Custom Scalar\n\nUse `scalar_type` decorator with a python class.\n\n```python\nfrom gql import scalar_type\n\n\n@scalar_type\nclass JSONString:\n    description = "The `JSONString` represents a json string."\n\n    @staticmethod\n    def serialize(value: Any) -> str:\n        return json.dumps(value)\n\n    @staticmethod\n    def parse_value(value: Any) -> dict:\n        if not isinstance(value, str):\n            raise TypeError(f\'JSONString cannot represent non string value: {inspect(value)}\')\n        return json.loads(value)\n\n    @staticmethod\n    def parse_literal(ast, _variables=None):\n        if isinstance(ast, StringValueNode):\n            return json.loads(ast.value)\n\n        return INVALID\n\n```\n\n\n## Apollo Federation\n\n[Example](https://github.com/syfun/starlette-graphql/tree/master/examples/federation)\n\n[Apollo Federation](https://www.apollographql.com/docs/apollo-server/federation/introduction/)\n\nThanks to [Ariadne](https://ariadnegraphql.org/docs/apollo-federation)\n\n\n## Framework support\n\n- [Starlette GraphQL](https://github.com/syfun/starlette-graphql)\n- [Django GraphQL](https://github.com/syfun/django-graphql)\n',
    'author': 'ysun',
    'author_email': 'sunyu418@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/syfun/python-gql',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
